// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of omnibus;

/// Handler pool base template
///
/// The handler pool manages handlers and funnel mapping. Handlers have to be
/// added and triggered through this class.
class _HandlerPool<T extends OmnibusHandler> {
  /// Handler functions
  List<T> _handlers = new List<T>();

  /// Funnel-less handlers
  Map<String, Set<int>> _funnelLess = new Map<String, Set<int>>();

  /// Pre-processed funnel map for handlers
  Map<String, _HandlerIndex> _funnelMap = new Map<String, _HandlerIndex>();

  /// Insert new handler
  void insert(T handler) {
    // Check if handler exists.
    if (handler != null) {
      // Insert new handler.
      int index = _handlers.length;
      _handlers.add(handler);

      // Check if the funnel contains items.
      if (handler.funnel != null && handler.funnel.isNotEmpty) {
        // Insert index in funnel map.
        _funnelMap.putIfAbsent(handler.omnibusId, () => new _HandlerIndex());
        _funnelMap[handler.omnibusId].insertIndex(index, handler.funnel);
      } else {
        // Insert handler in funnel-less handler list.
        _funnelLess.putIfAbsent(handler.omnibusId, () => new Set<int>());
        _funnelLess[handler.omnibusId].add(index);
      }
    }
  }

  /// Retieve handler indices that match the given funnel.
  Set<int> getIndices(String omnibusId, Map<String, dynamic> funnel) {
    // Include funnel-less handlers.
    Set<int> indices = _funnelLess.containsKey(omnibusId)
        ? _funnelLess[omnibusId]
        : new Set<int>();

    if (funnel != null && funnel.isNotEmpty) {
      // Check if a funnel index exists.
      if (_funnelMap.containsKey(omnibusId)) {
        return indices.union(_funnelMap[omnibusId].findIndices(funnel));
      } else {
        return indices;
      }
    } else {
      // Note: If the message data is empty this message could still be used as
      // a very simple trigger.
      return indices;
    }
  }
}

/// _HandlerPool for EventHandlers
class _EventHandlerPool extends _HandlerPool<OmnibusEventHandler> {
  /// Global event handlers
  List<GlobalEventHandler> globalHandlers = new List<GlobalEventHandler>();

  /// Publish
  ///
  /// This method runs in the future and all handlers are executed
  /// asynchronously.
  Future<List<Null>> publish(OmnibusEvent event) async {
    var indices = getIndices(event.omnibusId, event.getFunnel());
    var futures = new List<Future<Null>>();

    // Run global event handlers.
    for (final handler in globalHandlers) {
      futures.add(handler(event));
    }

    // Run event handlers.
    for (final i in indices) {
      futures.add(_handlers[i].process(event));
    }

    return Future.wait(futures);
  }
}

/// _HandlerPool for RequestHandlers
class _RequestHandlerPool extends _HandlerPool<OmnibusRequestHandler> {
  /// Global request handlers
  List<GlobalRequestHandler> globalHandlers = new List<GlobalRequestHandler>();

  /// Publish
  ///
  /// This method runs in the future and all handlers are executed
  /// asynchronously.
  Future<OmnibusResponses> publish(OmnibusRequest request) async {
    var indices = getIndices(request.omnibusId, request.getFunnel());
    var responses = new List<Future<OmnibusResponse>>();

    // Collect replies from global request handlers.
    for (final handler in globalHandlers) {
      responses.addAll(await handler(request));
    }

    // Collect replies.
    for (final i in indices) {
      responses.add(_handlers[i].process(request));
    }

    return new OmnibusResponses(responses);
  }
}
